# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for checking config_dict as hparams dictionary."""

import dataclasses
import json
import os
import sys

from absl import app
from absl import flags
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.utils import hparams_utils
from ml_collections import config_flags

FLAGS = flags.FLAGS

config_flags.DEFINE_config_file('hparams_config_dict', None,
                                'Path to file defining a config dict.')
flags.DEFINE_string(
    'output_dir',
    default=None,
    help='directory to dump hparams file to, if not dumps to stdout')


def main(_):
  if not FLAGS['hparams_config_dict'].present:
    raise ValueError(
        'Flag hparams_config_dict is required, but was not provided.')

  config = FLAGS.hparams_config_dict
  if 'configs' in config:
    config = config.configs[0]
    print('getting only first config dict')

  hparams = hparams_utils.load_hparams_from_config_dict(
      hparams_config.TrainingHParams, models.ResNet.HParams, config)

  data_dict = dataclasses.asdict(hparams)

  if FLAGS.output_dir is not None:
    output_path = os.path.join(FLAGS.output_dir, 'hparams_config.json')
    with open(output_path, 'w') as file:
      json.dump(data_dict, file, indent=2)
  else:
    json.dump(data_dict, sys.stdout, indent=2)


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util functions for training, which can be shared across models."""
from absl import logging
from aqt.jax_legacy.jax import quant_config


def should_quantize_weights(weight_quant_start_step: int, step: int) -> bool:
  return step >= weight_quant_start_step


def should_update_bounds(activation_bound_update_freq: int,
                         activation_bound_start_step: int, step: int) -> bool:
  """Returns whether activation bounds should be updated.

  Args:
    activation_bound_update_freq: How frequently to update bounds after the
      initial bounds update. A value of '-1' indicates to not update the bounds
      after the first update.
    activation_bound_start_step: The first step to update bounds on. '-1'
      indicates to never update bounds.
    step: The current training step.

  Returns:
    Boolean indicating whether to update the bounds on the current step.
  """
  if activation_bound_start_step < -1:
    raise ValueError("Start step must be >= -1.")
  if activation_bound_update_freq < -1 or activation_bound_update_freq == 0:
    raise ValueError("Update frequency must be a positive integer or -1.")
  steps_since_start = step - activation_bound_start_step
  if activation_bound_start_step == -1 or steps_since_start < 0:
    return False
  if activation_bound_update_freq == -1:
    return steps_since_start == 0
  else:
    return steps_since_start % activation_bound_update_freq == 0


def update_sparsity_mask(sparsity_start_step: int, sparsity_update_freq: int,
                         step: int) -> bool:
  """Returns whether sparsity mask should be updated.

  Args:
    sparsity_start_step: The first training step to start applying sparsity.
      Setting start step -1 indicates to not apply sparsity at any step.
    sparsity_update_freq: How frequently to update sparsity. Setting frequency 0
      indicates not to update again after the given step.
    step: The current training step.

  Returns:
    Boolean indicating whether to update the sparsity mask on the current step.
  """
  if sparsity_start_step < -1:
    raise ValueError("Start step must be >= -1.")
  if sparsity_update_freq < 0:
    raise ValueError("Update frequency must be a positive integer or 0.")
  steps_since_start = step - sparsity_start_step
  if sparsity_start_step == -1 or steps_since_start < 0:
    return False
  if sparsity_update_freq == 0:
    return steps_since_start == 0
  else:
    return steps_since_start % sparsity_update_freq == 0


# pylint: disable=g-doc-args
def get_dynamic_context_for_step(
    *,
    activation_bound_update_freq: int,
    activation_bound_start_step: int,
    step: int,
    collect_acts_stats: bool,
    prefer_int8_to_int32_dot: bool,
    sparsity_start_step: int,
    sparsity_update_freq: int,
) -> quant_config.DynamicContext:
  """Returns correct quantization context for a given step.

  Args:
    activation_bound_update_freq: How frequently to update bounds after the
      initial bounds update. A value of '-1' indicates to not update the bounds
      after the first update.
    activation_bound_start_step: The first step to update bounds on. '-1'
      indicates to never update bounds.
    step: The current training step.
    collect_acts_stats: Whether to collect activation statistics.
    prefer_int8_to_int32_dot: Whether to feed lax.dot inputs with an int8 dtype
      and accumulate to int32.

  Returns:
    A quant_config.DynamicContext instance.
  """
  update_bounds = should_update_bounds(
      activation_bound_start_step=activation_bound_start_step,
      activation_bound_update_freq=activation_bound_update_freq,
      step=step)
  apply_sparsity = False
  if sparsity_start_step >= 0:
    apply_sparsity = step >= sparsity_start_step
  update_weight_sparsity = update_sparsity_mask(sparsity_start_step,
                                                sparsity_update_freq, step)
  update_act_sparsity = update_sparsity_mask(sparsity_start_step,
                                             sparsity_update_freq, step)
  num_update_sparsity = 0.0
  if sparsity_update_freq == 0:
    logging.warning("Sparsity mask updates only once.")
  else:
    if step >= sparsity_start_step:
      num_update_sparsity = int(
          (step - sparsity_start_step) / sparsity_update_freq)
  # TODO(ayazdan): Relax this parameter to enable more iterations of decaying.
  num_update_sparsity = min(8, num_update_sparsity)
  quantize_acts = step >= activation_bound_start_step
  return quant_config.DynamicContext(
      update_bounds=update_bounds,
      quantize_acts=quantize_acts,
      apply_sparsity=apply_sparsity,
      update_weight_sparsity=update_weight_sparsity,
      update_act_sparsity=update_act_sparsity,
      num_update_sparsity=num_update_sparsity,
      collect_acts_stats=collect_acts_stats,
      prefer_int8_to_int32_dot=prefer_int8_to_int32_dot,
  )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collection of commonly used convenience functions for experiment analysis."""

import dataclasses
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple
from aqt.jax_legacy.utils import pandas_utils
from aqt.jax_legacy.utils import report_utils
import pandas as pd
import tree


def flatten_with_joined_string_paths(
    dictionary: Dict[str, Any]) -> Dict[str, Any]:
  """Flattens nested dict to single level dict with joined paths as keys."""
  flattened = tree.flatten_with_path(structure=dictionary)
  flattened_dict = {}
  # join path tuples to single string
  for path_tuple, val in flattened:
    # convert all path elements to strings
    path = [str(s) for s in path_tuple]
    path = '/'.join(path)
    flattened_dict[path] = val
  return flattened_dict


def convert_report_to_flat_dict_default(
    report: report_utils.ExperimentReport) -> Dict[str, Any]:
  """Selects subset of report and flattens it to a single level dict.

  This function selects all information except what's stored under the fields
  `report_query_args` and `metadata_corp`.

  This function serves as an example for how to parse an ExperimentReport
  into a dataframe row by flattening it to a row_dict, with keys corresponding
  to dataframe columns.

  The ExperimentReport dataclass likely contains more information then you need
  for your analysis, so you can write your own function to pick and choose the
  information you want. You can refer to report_utils.ExperimentReport for
  documentation of all available fields.

  You can pass your custom function into convert_reports_to_dataframe().

  Args:
    report: An instance of ExperimentReport.

  Returns:
    A flattened dict representing a dataframe row.
  """

  row_dict = {}

  # Add smoothed metrics if present
  if report.metrics is not None:
    flattened_metrics = dict(flatten_with_joined_string_paths(report.metrics))
    # merge dicts
    row_dict = {**row_dict, **flattened_metrics}

  # Add unsmoothed metrics if present
  if report.unsmoothed_metrics is not None:
    flattened_unsmoothed_metrics = dict(
        flatten_with_joined_string_paths(report.unsmoothed_metrics))
    flattened_unsmoothed_metrics = {
        f'unsmoothed/{k}': v for k, v in flattened_unsmoothed_metrics.items()
    }
    # merge dicts
    row_dict = {**row_dict, **flattened_unsmoothed_metrics}


  # Ignore following fields because they have already been added, or we chose
  # not to include them.
  report_fields_to_ignore = {
      'metrics',
      'unsmoothed_metrics',
  }
  # Add other report fields.
  for field in dataclasses.fields(report):
    if field.name not in report_fields_to_ignore:
      row_dict[field.name] = getattr(report, field.name)

  return row_dict


def convert_reports_to_dataframe(
    reports: List[report_utils.ExperimentReport],
    convert_report_to_flat_dict_fn: Callable[
        [report_utils.ExperimentReport],
        Dict[str, Any]] = convert_report_to_flat_dict_default
) -> pd.DataFrame:
  """Converts a list of ExperimentReport instances to a pandas dataframe.

  Args:
    reports: List of ExperimentReport instances. Each instance will correspond
      to a row in the dataframe.
    convert_report_to_flat_dict_fn: Function to use for converting an
      ExperimentReport to a flat dict, which will then be read in as a pandas
      dataframe row. The keys in the flat dict are interpreted as column names,
      the values as entries for that row. Please refer to
      `convert_report_to_flat_dict_default()` as an example.

  Returns:
    A pandas dataframe populated with information extracted from the reports.
  """

  rows = [convert_report_to_flat_dict_fn(rep) for rep in reports]
  return pd.DataFrame(rows)


def clickable_link(link: str, display_str: str = 'link') -> str:
  """Converts a link string into a clickable link with html tag.

  WARNING: This function is not safe to use for untrusted inputs since the
  generated HTML is not sanitized.

  Usage:
    df.style.format(clickable_link, subset=['col_name'])

  Args:
    link: A link string without formatting.
    display_str: What text the link should display.

  Returns:
    HTML-formatted link.

  """
  return f'<a href="{link}">{display_str}</a>'





# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT tensor."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import tiled_dot_general
import jax
import jax.numpy as jnp


TensorTiling = tiled_dot_general.TensorTiling
AxisTiling = tiled_dot_general.AxisTiling


# TODO(yichizh): refactor and clean the test cases
class AqtTensorTest(parameterized.TestCase):

  def test_dynamic_slice(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=1, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)

    q = aqt_tensor.QTensor(
        qvalue=x,
        scale=[scale],
        scale_t=None,
        bias=[],
        dequant_dtype=scale.dtype,
    )
    y = aqt_tensor.dynamic_slice(q, start_indices=(1, 0), slice_sizes=[2, 1])
    print("======")
    print(y.qvalue)
    print(y.scale[0])

  def test_getitem(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=0, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)

    q = aqt_tensor.QTensor(
        qvalue=x,
        scale=[scale],
        scale_t=None,
        bias=[],
        dequant_dtype=scale.dtype,
    )
    y = q.__getitem__(2)
    print("======")
    print(y.qvalue)
    print(y.scale[0])

  def test_dynamic_update(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=0, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)
    q = aqt_tensor.QTensor(
        qvalue=x,
        scale=[scale],
        scale_t=None,
        bias=[],
        dequant_dtype=scale.dtype,
    )

    update_qvalue = jnp.zeros((3, 1), dtype=x.dtype)
    update_scale = jnp.max(jnp.abs(update_qvalue), axis=0, keepdims=True)
    update = aqt_tensor.QTensor(
        qvalue=update_qvalue,
        scale=[update_scale],
        scale_t=None,
        bias=[],
        dequant_dtype=update_scale.dtype,
    )
    y = aqt_tensor.dynamic_update_slice(q, update, (0, 1))
    print("======")
    print(y.qvalue)
    print(y.scale[0])

  def test_dtype(self):
    qt = aqt_tensor.zeros(
        shape=(1,), container_dtype=jnp.int8, dequant_dtype=jnp.float32
    )
    self.assertEqual(qt.dtype, jnp.float32)
    self.assertIsInstance(qt.dtype, jnp.dtype)
    self.assertEqual(qt.dequant_dtype, jnp.float32)

    qt = qt.astype(jnp.bfloat16)
    self.assertEqual(qt.dtype, jnp.bfloat16)
    self.assertIsInstance(qt.dtype, jnp.dtype)
    self.assertEqual(qt.dequant_dtype, jnp.bfloat16)

  def test_tiling_state(self):
    lhs_key, rhs_key = jax.random.split(jax.random.PRNGKey(0))
    lhs = jax.random.randint(lhs_key, shape=(30, 40, 50), minval=0, maxval=100)
    rhs = jax.random.randint(rhs_key, shape=(40, 50, 60), minval=0, maxval=100)

    cfg = tiled_dot_general.Cfg(
        lhs=TensorTiling(
            contraction_axes=[
                AxisTiling(axis=1, tile_size=10, tile_count=None),
                AxisTiling(axis=2, tile_size=25, tile_count=None),
            ],
            remaining_axes=[],
        ),
        rhs=TensorTiling(
            contraction_axes=[
                AxisTiling(axis=0, tile_size=10, tile_count=None),
                AxisTiling(axis=1, tile_size=25, tile_count=None),
            ],
            remaining_axes=[],
        ),
    )

    # make tiling state for both lhs and rhs
    lhs_ca = (1, 2)
    rhs_ca = (0, 1)
    xlhs, xrhs = tiled_dot_general.generate_tiling_states_for_dot_general(
        cfg, lhs, rhs, dimension_numbers=((lhs_ca, rhs_ca), ((), ()))
    )

    # make scale
    _, xlhs_ca = xlhs.to_tiled_axes_transposed(lhs_ca)
    xlhs_scale = xlhs.apply(lhs)
    xlhs_scale = jnp.max(xlhs_scale, axis=xlhs_ca, keepdims=True) / (2.0**7 - 1)

    _, xrhs_ca = xrhs.to_tiled_axes_transposed(rhs_ca)
    xrhs_scale = xrhs.apply(rhs)
    xrhs_scale = jnp.max(xrhs_scale, axis=xrhs_ca, keepdims=True) / (2.0**7 - 1)

    # make qtensor
    qlhs = aqt_tensor.QTensor(
        qvalue=None,
        scale=[xlhs_scale],
        scale_t=None,
        bias=[],
        dequant_dtype=xlhs_scale.dtype,
        tiling_state=xlhs,
    )
    qlhs = qlhs.quant(lhs)

    qrhs = aqt_tensor.QTensor(
        qvalue=None,
        scale=[xrhs_scale],
        scale_t=None,
        bias=[],
        dequant_dtype=xrhs_scale.dtype,
        tiling_state=xrhs,
    )
    qrhs = qrhs.quant(rhs)

    self.assertEqual(qlhs.shape, lhs.shape)
    self.assertEqual(qrhs.shape, rhs.shape)
    self.assertEqual(qlhs.ndim, lhs.ndim)
    self.assertEqual(qrhs.ndim, rhs.ndim)

    self.assertEqual(qlhs.qvalue.shape, tuple(xlhs.tiled_shape))
    self.assertEqual(qrhs.qvalue.shape, tuple(xrhs.tiled_shape))

    self.assertEqual(qlhs.dequant().shape, lhs.shape)
    self.assertEqual(qrhs.dequant().shape, rhs.shape)

    # Check assertion raised when wrong tiling_state is given.
    qlhs.tiling_state.tiled_shape = qrhs.qvalue.shape
    with self.assertRaises(AssertionError):
      qlhs.dequant()
    with self.assertRaises(AssertionError):
      qlhs.shape  # pylint: disable=pointless-statement

    qrhs.tiling_state.tiled_shape = qlhs.qvalue.shape
    with self.assertRaises(AssertionError):
      qrhs.dequant()
    with self.assertRaises(AssertionError):
      qrhs.shape  # pylint: disable=pointless-statement


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General dot_general tiling."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import copy
import dataclasses
import pprint
from typing import Iterable
from absl import logging
from aqt.jax.v2 import utils
import jax
import jax.numpy as jnp
from typing_extensions import Self  # for python version < 3.11


AxisIdx = utils.AxisIdx
AxisSize = utils.AxisSize
AqtTileMap = dict[AxisIdx | str, list[AxisIdx]]
EinsumEqnLetter = str
EinsumTileSizes = dict[EinsumEqnLetter, AxisSize]

BROADCAST_PREFIX = 'broadcast_'


@dataclasses.dataclass(frozen=False, slots=True)
class AxisTiling:
  """Axis tiling configuration for subchannel quantization."""
  axis: AxisIdx
  # At most one of tile_count, tile_size can be None.
  tile_count: None | AxisSize = None
  tile_size: None | AxisSize = None

  def __post_init__(self):
    msg = 'At most one of tile_count and tile_size can be None'
    assert not (self.tile_count is None and self.tile_size is None), msg

  def complete_missing(self, shape):
    """Completes missing tile_count or tile_size."""
    if self.axis < 0:
      # make nagative index to positive index
      self.axis = len(shape) + self.axis
    tc = self.tile_count
    ts = self.tile_size
    axis_size = shape[self.axis]
    msg = 'At most one of tile_count and tile_size can be None'
    assert not (tc is None and ts is None), msg
    self.tile_count = axis_size // ts if tc is None else tc
    self.tile_size = axis_size // tc if ts is None else ts
    msg = (
        f'Axis {self.axis} cannot be split into'
        f' {self.tile_count} tiles.'
    )
    assert self.tile_size * self.tile_count == axis_size, msg


@dataclasses.dataclass(frozen=False, slots=True)
class TensorTiling:
  contraction_axes: list[AxisTiling]
  remaining_axes: list[AxisTiling]
  # There is no point in tiling dot_general's batch axes


@dataclasses.dataclass(frozen=False, slots=True)
class Cfg:
  """Sequence of (lhs, rhs) configurations."""
  lhs: TensorTiling
  rhs: TensorTiling

  @classmethod
  def from_einsum(cls, eqn: str, einsum_tile_sizes: EinsumTileSizes) -> Self:
    """Creates Cfg based on einsum equation and tile sizes."""
    args_eqn, ret_eqn1 = eqn.split('->')
    args_eqn = args_eqn.split(',')
    assert len(args_eqn) == 2, f'more then 2 arguments are not supported {eqn}'
    lhs_eqn1, rhs_eqn1 = args_eqn
    assert lhs_eqn1.isalpha(), f'unsupported lhs: {eqn}'
    assert rhs_eqn1.isalpha(), f'unsupported rhs: {eqn}'
    assert ret_eqn1.isalpha(), f'unsupported ret: {eqn}'
    ret = Cfg(
        lhs=TensorTiling(contraction_axes=[], remaining_axes=[]),
        rhs=TensorTiling(contraction_axes=[], remaining_axes=[]),
    )
    for einsum_letter, tile_size in einsum_tile_sizes.items():
      assert einsum_letter.isalpha()
      assert len(einsum_letter) == 1
      in_lhs = lhs_eqn1.find(einsum_letter)
      in_rhs = rhs_eqn1.find(einsum_letter)
      in_ret = ret_eqn1.find(einsum_letter)
      found_in_lhs = in_lhs != -1
      found_in_rhs = in_rhs != -1
      found_in_ret = in_ret != -1

      def make_tiling(axis, tile_size):
        return AxisTiling(axis=axis, tile_size=tile_size, tile_count=None)

      msg = (
          'We support only contraction axes and remaining axes:'
          f' eqn={eqn} einsum_letter={einsum_letter}'
      )
      assert found_in_lhs + found_in_rhs + found_in_ret == 2, msg
      if found_in_lhs and found_in_rhs:
        # Contraction axis
        ret.lhs.contraction_axes.append(make_tiling(in_lhs, tile_size))
        ret.rhs.contraction_axes.append(make_tiling(in_rhs, tile_size))
      else:
        # Remaining axis
        assert found_in_ret, f'Should not happen. {msg}'
        if found_in_lhs:
          ret.lhs.remaining_axes.append(make_tiling(in_lhs, tile_size))
        if found_in_rhs:
          ret.rhs.remaining_axes.append(make_tiling(in_rhs, tile_size))
    return ret

  def complete_missing(
      self,
      lhs_shape: tuple[AxisSize, ...],
      rhs_shape: tuple[AxisSize, ...],
  ) -> Self:
    """Makes lhs and rhs to cover all the axes."""
    new_cfg = copy.deepcopy(self)

    for ax in (*new_cfg.lhs.contraction_axes, *new_cfg.lhs.remaining_axes):
      ax.complete_missing(lhs_shape)
    for ax in (*new_cfg.rhs.contraction_axes, *new_cfg.rhs.remaining_axes):
      ax.complete_missing(rhs_shape)
    return new_cfg


def interleave(tile_count, tile_size, tile_map, remaining_axes, product=False):
  """Interleave the tile_count and tile_size of remaining axes."""
  tile_count = list(tile_count)
  tile_size = list(tile_size)
  ret = []
  # For each tile size, check if the length of the tile_map is 1 or 2.
  # If 1, the axis is not tiled. Directly append the tile size to ret.
  # If 2, the axis is tiled. Populate a tile count and append it to ret.
  for i in range(len(tile_size)):
    ts = tile_size[i]
    if len(tile_map[remaining_axes[i]]) > 1:
      tc = tile_count.pop(0)
    else:
      tc = None  # None means not tiled. It is different from tiled but tc=1.
    if product:
      ret.append(ts if tc is None else tc * ts)
    else:
      ret.extend((ts,) if tc is None else (tc, ts))
  msg = 'Remaining axes tile count and tile size are not fully interleaved.'
  assert len(tile_count) == 0, msg
  return ret


def get_ra(rank, ca, ba) -> list[AxisIdx]:
  return list(a for a in range(rank) if a not in ca + ba)


def maybe_add_one(i, min_i):
  return i + int(i >= min_i)


@dataclasses.dataclass(frozen=False, slots=True)
class TilingState:
  """Structure for bookkeeping of AxisIdx while tiling."""

  untiled_shape: tuple[AxisSize, ...]

  # Below are axes indices, similar to dimension_numbers
  # E.g.
  #   assert xhs.shape == [10, 20, 30, ...],
  #   assert tile_map == {0:[0], 1:[1], 2:[2], ...}
  # After splitting axis 1 to 2 tiles, sized 10, we get
  #   assert xhs.shape == [10, 2, 10, 30, ...]
  #   assert tile_map == {0:[0], 1:[1,2], 2:[3], ...}
  tile_map: AqtTileMap = utils.dataclass_field(dict)

  tiled_shape: list[AxisSize] = utils.dataclass_field(list)

  def __post_init__(self):
    for i in range(len(self.untiled_shape)):
      # No axis is split at all yet.
      self.tile_map[i] = [i]
    self.tiled_shape = list(self.untiled_shape)

  def tile_one_axis(self, at: AxisTiling):
    """Tiles (splits) one axis while maintaining all AxisIdx."""
    msg = "Can't tile as all tiling must be done before broadcast operations."
    assert len(self.get_broadcasted_tile_map_indexes()) == 0, msg

    tile_axis = self.tile_map[at.axis]
    assert len(tile_axis) == 1, "can't tile the same axis twice."
    tile_axis = tile_axis[0]

    msg = f'{self.tiled_shape[tile_axis]=}, {at.tile_size=}, {at.tile_count=}'
    assert self.tiled_shape[tile_axis] == at.tile_size * at.tile_count, msg
    self.tiled_shape[tile_axis] = at.tile_size
    self.tiled_shape.insert(tile_axis, at.tile_count)

    # Update tile_map
    for k in self.tile_map:
      self.tile_map[k] = [
          maybe_add_one(ai, tile_axis) for ai in self.tile_map[k]
      ]
    self.tile_map[at.axis] = [tile_axis, tile_axis + 1]

  def tile_axes(self, ats: Iterable[AxisTiling]):
    for at in ats:
      self.tile_one_axis(at)

  def _is_broadcasted_ax(self, tile_map_idx: AxisIdx | str) -> bool:
    return isinstance(tile_map_idx, str) and tile_map_idx.startswith(
        BROADCAST_PREFIX
    )

  def get_broadcasted_tile_map_indexes(self):
    """Returns the list of keys in `self.tile_map` associated with broadcasting axes."""
    return [ax for ax in self.tile_map if self._is_broadcasted_ax(ax)]

  def apply(self, x: jnp.ndarray) -> jnp.ndarray:
    num_bcast_axes = len(self.get_broadcasted_tile_map_indexes())
    tiled_x = x.reshape(self.tiled_shape[num_bcast_axes:])
    tiled_x = jnp.broadcast_to(tiled_x, self.tiled_shape)
    return tiled_x

  def unapply(self, tiled_x: jnp.ndarray) -> jnp.ndarray:
    num_bcast_axes = len(self.get_broadcasted_tile_map_indexes())
    # All elements of broadcast axes are the same, thus we take the first one.
    first_index = (0,) * num_bcast_axes
    x = tiled_x[jnp.s_[first_index]]
    x = x.reshape(self.untiled_shape)
    return x

  def broadcast_to_other(self, bcast_shape: tuple[AxisSize, ...]):
    """Adds new axes (bcast_shape) on AxisIdx=0."""
    for k in self.tile_map:
      self.tile_map[k] = [ai + len(bcast_shape) for ai in self.tile_map[k]]
    keys = []
    for i in range(len(bcast_shape)):
      k = f'{BROADCAST_PREFIX}{i}'
      keys.append(k)
      assert k not in self.tile_map
      self.tile_map[k] = [i]
    self.tiled_shape = list(bcast_shape) + self.tiled_shape
    return keys

  def axes_shape(self, axes: list[AxisIdx]) -> tuple[AxisSize, ...]:
    return tuple(map(lambda a: self.tiled_shape[a], axes))

  def to_tiled_axes_transposed(
      self,
      axes: Iterable[AxisIdx | str],
  ) -> tuple[list[AxisIdx], None | list[AxisIdx]]:
    # pylint: disable=g-doc-args,g-doc-return-or-yield
    """The given 'axes' parameter defines axes in untiled represented Array.

    Function returns the corresponding tile count and tile size AxisIdxs in
    self.x which is tiled.
    The return value depends on the length of the tile_map[ai].
    If the length is 2, 1st element is tile count and 2nd element is tile size
    If the length is 1, the only element is tile size.
    """
    tile_count = []
    tile_size = []
    for ai in axes:
      if isinstance(ai, AxisIdx) and ai < 0:
        # make nagative index to positive index
        ai = len(self.untiled_shape) + ai
      tiled = self.tile_map[ai]
      match len(tiled):
        case 1:
          tile_size.append(tiled[0])
        case 2:
          tile_count.append(tiled[0])
          tile_size.append(tiled[1])
    return tile_count, tile_size


def print_dimension_numbers(dimension_numbers, lhs, rhs, label) -> None:
  """Prints dimension numbers before and/or after tiling.

  Args:
    dimension_numbers: It contains the left and right hand side contraction axes
      and batch axes.
    lhs: left hand side tensor.
    rhs: right hand side tensor.
    label: A string tag for logging info.
  """
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  lhs_ra = get_ra(lhs.ndim, lhs_ca, lhs_ba)
  rhs_ra = get_ra(rhs.ndim, rhs_ca, rhs_ba)
  logging.vlog(1, label)
  logging.vlog(1, f' lhs.shape={lhs.shape}')
  logging.vlog(1, f' rhs.shape={rhs.shape}')
  logging.vlog(1, f'lhs_ca={lhs_ca}')
  logging.vlog(1, f'rhs_ca={rhs_ca}')
  logging.vlog(1, f'lhs_ba={lhs_ba}')
  logging.vlog(1, f'rhs_ba={rhs_ba}')
  logging.vlog(1, f'lhs_ra={lhs_ra}')
  logging.vlog(1, f'rhs_ra={rhs_ra}')


def generate_tiling_state(
    tensor: jnp.ndarray,
    tiled_axes: Iterable[AxisTiling],
) -> TilingState:
  """Generates tiling states for the given tensor."""
  for ax in tiled_axes:
    ax.complete_missing(tensor.shape)

  xtensor = TilingState(untiled_shape=tensor.shape)
  xtensor.tile_axes(tiled_axes)
  return xtensor


def generate_tiling_states_for_dot_general(
    cfg: Cfg,
    lhs,
    rhs,
    dimension_numbers: jax.lax.DotDimensionNumbers,
) -> tuple[TilingState, TilingState]:
  """Do tiling for `lhs` and `rhs` and returns the intermediate tiling states.

  Args:
    cfg: Configuration to define how tiling should be done
    lhs: The left operand of dot general
    rhs: The right operand of dot general
    dimension_numbers: dimension numbers to indicate batch dimensions and
      contracting dimensions

  Returns:
    xlhs: Tiling state of the left operand `lhs` after the tiling
    xrhs: Tiling state of the left operand `rhs` after the tiling
  """
  logging.vlog(1, 'Tiling config cfg: %s', cfg)
  print_dimension_numbers(dimension_numbers, lhs, rhs, label='before tiling')

  # Config pre-processing and verification

  cfg = copy.deepcopy(cfg)
  cfg = cfg.complete_missing(lhs.shape, rhs.shape)
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  lhs_ra = get_ra(lhs.ndim, lhs_ca, lhs_ba)
  rhs_ra = get_ra(rhs.ndim, rhs_ca, rhs_ba)
  g_msg = (
      'Before tiling: \n'
      f'lhs: {lhs.shape=}, {lhs_ca=}, {lhs_ba=}, {lhs_ra=} \n'
      f'rhs: {rhs.shape=}, {rhs_ca=}, {rhs_ba=}, {rhs_ra=} \n'
      f'tiling cfg: {pprint.pformat(cfg)} \n'
  )

  # First tile_axis CA. CA tile_count axes will be first in xxhs.ba_tile
  assert len(cfg.lhs.contraction_axes) == len(cfg.rhs.contraction_axes), g_msg
  for cfg_lhs_ca, cfg_rhs_ca in zip(
      cfg.lhs.contraction_axes, cfg.rhs.contraction_axes
  ):
    msg = (
        'Contraction axis tile counts should be the same, but found lhs axis'
        f' {cfg_lhs_ca.axis} has a tile count of {cfg_lhs_ca.tile_count}, and'
        f' rhs axis {cfg_rhs_ca.axis} has a tile count of'
        f' {cfg_rhs_ca.tile_count}'
    )
    assert cfg_lhs_ca.tile_count == cfg_rhs_ca.tile_count, msg

  # Tiling

  xlhs = TilingState(untiled_shape=lhs.shape)
  xlhs.tile_axes(cfg.lhs.contraction_axes + cfg.lhs.remaining_axes)
  xrhs = TilingState(untiled_shape=rhs.shape)
  xrhs.tile_axes(cfg.rhs.contraction_axes + cfg.rhs.remaining_axes)

  # DotGeneral reshapeing

  xlhs_ra_tile, _ = xlhs.to_tiled_axes_transposed(lhs_ra)
  xrhs.broadcast_to_other(xlhs.axes_shape(xlhs_ra_tile))

  xrhs_ra_tile, _ = xrhs.to_tiled_axes_transposed(rhs_ra)
  xlhs.broadcast_to_other(xrhs.axes_shape(xrhs_ra_tile))

  return xlhs, xrhs


def tiled_dot_general_with_tiling_states(
    lhs: jnp.ndarray,
    xlhs: TilingState,
    rhs: jnp.ndarray,
    xrhs: TilingState,
    untiled_dimension_numbers: jax.lax.DotDimensionNumbers,
    precision=None,
    preferred_element_type=None,
    dot_general=jax.lax.dot_general,
):
  """local dot_general with tiling states."""
  xlhs_x = xlhs.apply(lhs)
  xrhs_x = xrhs.apply(rhs)
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = untiled_dimension_numbers
  lhs_ra = get_ra(lhs.ndim, lhs_ca, lhs_ba)
  rhs_ra = get_ra(rhs.ndim, rhs_ca, rhs_ba)
  xlhs_ca_tile, xlhs_ca = xlhs.to_tiled_axes_transposed(lhs_ca)
  xlhs_ra_tile, xlhs_ra = xlhs.to_tiled_axes_transposed(lhs_ra)
  xrhs_ca_tile, xrhs_ca = xrhs.to_tiled_axes_transposed(rhs_ca)
  xrhs_ra_tile, xrhs_ra = xrhs.to_tiled_axes_transposed(rhs_ra)
  xlhs_bcast = xlhs.get_broadcasted_tile_map_indexes()
  xrhs_bcast = xrhs.get_broadcasted_tile_map_indexes()
  # We can pattern match against [] because batch axes can't be tiled and
  # all of them will be returned as tile sizes.
  [], xlhs_ra_tile_other = xlhs.to_tiled_axes_transposed(xlhs_bcast)
  [], xrhs_ra_tile_other = xrhs.to_tiled_axes_transposed(xrhs_bcast)
  [], xlhs_ba = xlhs.to_tiled_axes_transposed(lhs_ba)
  [], xrhs_ba = xrhs.to_tiled_axes_transposed(rhs_ba)

  tiled_ca = (xlhs_ca, xrhs_ca)
  tiled_ba = (
      xlhs_ca_tile + xlhs_ba + xlhs_ra_tile + xlhs_ra_tile_other,
      xrhs_ca_tile + xrhs_ba + xrhs_ra_tile_other + xrhs_ra_tile,
  )
  tiled_dimension_numbers = (tiled_ca, tiled_ba)

  tiled_lhs_ra = get_ra(xlhs_x.ndim, tiled_ca[0], tiled_ba[0])
  tiled_rhs_ra = get_ra(xrhs_x.ndim, tiled_ca[1], tiled_ba[1])
  g_msg = (
      'After tiling: \n'
      f' lhs: lhs.shape={xlhs_x.shape}, lhs_ca={tiled_ca[0]},'
      f' lhs_ba={tiled_ba[0]}, lhs_ra={tiled_lhs_ra} \n'
      f' rhs: rhs.shape={xrhs_x.shape}, rhs_ca={tiled_ca[1]},'
      f' rhs_ba={tiled_ba[1]}, rhs_ra={tiled_rhs_ra} \n'
  )
  for axis in tiled_ca[0] + tiled_ba[0]:
    assert axis >= 0 and axis < xlhs_x.ndim, g_msg
  for axis in tiled_ca[1] + tiled_ba[1]:
    assert axis >= 0 and axis < xrhs_x.ndim, g_msg

  out = dot_general(
      xlhs_x, xrhs_x, tiled_dimension_numbers, precision, preferred_element_type
  )

  print_dimension_numbers(
      tiled_dimension_numbers, xlhs_x, xrhs_x, label='after tiling'
  )

  # Some assertions
  assert xlhs.axes_shape(xlhs_ca_tile) == xrhs.axes_shape(xrhs_ca_tile), g_msg
  ca_tile_sh = xlhs.axes_shape(xlhs_ca_tile)

  assert xlhs.axes_shape(xlhs_ba) == xrhs.axes_shape(xrhs_ba), g_msg
  ba_sh = xlhs.axes_shape(xlhs_ba)

  assert xlhs.axes_shape(xlhs_ra_tile) == xrhs.axes_shape(
      xrhs_ra_tile_other
  ), g_msg
  lhs_ra_tile_sh = xlhs.axes_shape(xlhs_ra_tile)

  assert xlhs.axes_shape(xlhs_ra_tile_other) == xrhs.axes_shape(
      xrhs_ra_tile
  ), g_msg
  rhs_ra_tile_sh = xlhs.axes_shape(xlhs_ra_tile_other)

  lhs_ra_sh = xlhs.axes_shape(xlhs_ra)
  rhs_ra_sh = xrhs.axes_shape(xrhs_ra)

  g_msg += f'Tiled dg {out.shape=} \n'
  assert (
      out.shape
      == ca_tile_sh
      + ba_sh
      + lhs_ra_tile_sh
      + rhs_ra_tile_sh
      + lhs_ra_sh
      + rhs_ra_sh
  ), g_msg

  # Sum over ca_tile now.
  # all_ba() returns ca_tile as the first axes in ba.
  assert len(xlhs_ca_tile) == len(xrhs_ca_tile)
  out = out.sum(axis=range(len(xlhs_ca_tile)))

  g_msg += f'After sum over tiles {out.shape=} \n'
  assert (
      out.shape
      == ba_sh + lhs_ra_tile_sh + rhs_ra_tile_sh + lhs_ra_sh + rhs_ra_sh
  ), g_msg

  # Transpose tile and tile size together
  # Axis tracking is obsolete here. Only the length is the same.
  start = 0
  end_ba = len(ba_sh)
  end_lhs_ra_tile = end_ba + len(lhs_ra_tile_sh)
  end_rhs_ra_tile = end_lhs_ra_tile + len(rhs_ra_tile_sh)
  end_lhs_ra = end_rhs_ra_tile + len(lhs_ra_sh)
  end_rhs_ra = end_lhs_ra + len(rhs_ra_sh)

  from_to = lambda end1, end2: list(range(end1, end2))

  new_ba = from_to(start, end_ba)
  new_lhs_ra_tile = from_to(end_ba, end_lhs_ra_tile)
  new_rhs_ra_tile = from_to(end_lhs_ra_tile, end_rhs_ra_tile)
  new_lhs_ra = from_to(end_rhs_ra_tile, end_lhs_ra)
  new_rhs_ra = from_to(end_lhs_ra, end_rhs_ra)

  out = out.transpose(
      new_ba
      + interleave(new_lhs_ra_tile, new_lhs_ra, xlhs.tile_map, lhs_ra)
      + interleave(new_rhs_ra_tile, new_rhs_ra, xrhs.tile_map, rhs_ra)
  )

  lhs_ra_sh_interleaved = tuple(
      interleave(lhs_ra_tile_sh, lhs_ra_sh, xlhs.tile_map, lhs_ra)
  )
  rhs_ra_sh_interleaved = tuple(
      interleave(rhs_ra_tile_sh, rhs_ra_sh, xrhs.tile_map, rhs_ra)
  )
  lhs_ra_sh_flattened = tuple(
      interleave(lhs_ra_tile_sh, lhs_ra_sh, xlhs.tile_map, lhs_ra, product=True)
  )
  rhs_ra_sh_flattened = tuple(
      interleave(rhs_ra_tile_sh, rhs_ra_sh, xrhs.tile_map, rhs_ra, product=True)
  )

  g_msg += f'After transpose {out.shape=} \n'
  assert (
      out.shape == ba_sh + lhs_ra_sh_interleaved + rhs_ra_sh_interleaved
  ), g_msg
  out = out.reshape(ba_sh + lhs_ra_sh_flattened + rhs_ra_sh_flattened)

  return out


def tiled_dot_general(
    cfg: Cfg,
    lhs,
    rhs,
    dimension_numbers,
    precision=None,
    preferred_element_type=None,
    dot_general=jax.lax.dot_general,
):
  """local dot_general."""
  xlhs, xrhs = generate_tiling_states_for_dot_general(
      cfg, lhs, rhs, dimension_numbers
  )
  return tiled_dot_general_with_tiling_states(
      lhs,
      xlhs,
      rhs,
      xrhs,
      dimension_numbers,
      precision,
      preferred_element_type,
      dot_general,
  )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT flax intercept methods."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2.flax import aqt_flax
from aqt.jax.v2.flax.intercept import aqt_intercept_methods
import flax.linen as nn
import jax
import numpy as np


class MlpBlock(nn.Module):
  @nn.compact
  def __call__(self, inputs):
    x = nn.Dense(features=inputs.shape[-1] * 4)(inputs)
    x = nn.relu(x)
    x = nn.Dense(features=inputs.shape[-1])(x)
    return x


class NestedMlpBlock(nn.Module):
  @nn.compact
  def __call__(self, inputs):
    x = MlpBlock()(inputs)
    x = nn.relu(x)
    x = MlpBlock()(x)
    return x


class MockDotGeneralGenerator(aqt_intercept_methods.DotGeneralGenerator):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.count = 0

  def __call__(self, *args, **kwargs):
    self.count += 1
    return super().__call__(*args, **kwargs)


class MockDotGeneralGeneratorByModule(
    aqt_intercept_methods.DotGeneralGeneratorByModule):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.count = 0

  def generate_by_module(self, module: nn.Module):
    if isinstance(module, nn.Dense):
      self.count += 1
    return None


class AqtInterceptMethodsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(testcase_name="dot_general_generator",
           model_cls=MlpBlock,
           dot_general_generator_cls=MockDotGeneralGenerator,
           expected_count=3),
      dict(testcase_name="dot_general_generator_by_module",
           model_cls=MlpBlock,
           dot_general_generator_cls=MockDotGeneralGeneratorByModule,
           expected_count=2),
      dict(testcase_name="dot_general_generator_nested",
           model_cls=NestedMlpBlock,
           dot_general_generator_cls=MockDotGeneralGenerator,
           expected_count=7),
      dict(testcase_name="dot_general_generator_by_module_nested",
           model_cls=NestedMlpBlock,
           dot_general_generator_cls=MockDotGeneralGeneratorByModule,
           expected_count=4))
  def test_intercept_methods_replace_dot_general_count(
      self, model_cls, dot_general_generator_cls, expected_count):
    np_seed = 0
    init_seed = 0
    eval_seed = 0
    mlp_block = model_cls()
    np.random.seed(np_seed)
    inputs = np.random.normal(size=(3, 4))
    mock_dot_general_generator = dot_general_generator_cls(
        dot_general=jax.lax.dot_general)
    with aqt_intercept_methods.intercept_methods_replace_dot_general(
        mock_dot_general_generator):
      model = mlp_block.init(jax.random.PRNGKey(init_seed), inputs)
      _ = mlp_block.apply(
          model, inputs, rngs={"params": jax.random.key(eval_seed)})

    self.assertEqual(mock_dot_general_generator.count, expected_count)

  def test_aqt_dot_general_generator(self):
    aqt_dot_general_generator = (
        aqt_intercept_methods.AqtDotGeneralGenerator(config.config_v4())
    )
    # Only support Dense.
    self.assertIsInstance(
        aqt_dot_general_generator.generate_by_module(nn.Dense(10)),
        aqt_flax.AqtDotGeneral)
    self.assertIsNone(
        aqt_dot_general_generator.generate_by_module(
            nn.Conv(10, (3, 3))))
    self.assertIsNone(
        aqt_dot_general_generator.generate_by_module(
            nn.Einsum((10, 10), "ab,bc->ac")))


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A sample to show how to calibrate and quantize a CNN model using AQT."""

from collections.abc import Sequence
import functools

from absl import app
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2.examples.cnn import aqt_utils
from aqt.jax.v2.examples.cnn import cnn_model
from aqt.jax.v2.examples.cnn import model_utils
from aqt.jax.v2.flax import aqt_flax_calibration
import jax


def _set_static_range_calib_options(aqt_cfg: aqt_config.DotGeneral) -> None:
  sr_calibration_cls = functools.partial(
      aqt_flax_calibration.MeanOfAbsMaxCalibration, quant_collection='qc'
  )
  aqt_config.set_fwd_calibration(aqt_cfg, sr_calibration_cls)
  aqt_cfg.fwd.dg_quantizer.lhs.calib_shared_axes = 'per_tensor'


def run(train_ds: dict[str, jax.Array], test_ds: dict[str, jax.Array]) -> None:
  """Extracted main function for unit testing."""
  # 1. Train
  state = model_utils.create_train_state(
      jax.random.key(0), cnn_model.CNN(True), cnn_model.CNN(False)
  )
  state = model_utils.train_and_evaluate(
      num_epochs=1,
      workdir='/tmp/aqt_mnist_example',
      train_ds=train_ds,
      test_ds=test_ds,
      state=state,
  )

  # 2. Calibrate
  aqt_cfg = aqt_config.config_v4(fwd_bits=8)
  _set_static_range_calib_options(aqt_cfg)
  state = aqt_utils.calibrate(cnn_model.CNN, aqt_cfg, state, 10, train_ds)

  # 3. Quantize and test
  loss = aqt_utils.serve_quantized(
      cnn_model.CNN, test_ds, aqt_cfg, state.model_vars, act_calibrated=True
  )
  print('serve loss on sample ds: {}'.format(loss))


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  train_ds, test_ds = cnn_model.get_datasets()
  run(train_ds, test_ds)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for model training and serving."""

from collections.abc import Callable
import functools
import sys
from typing import Any

from flax import linen as nn
from flax import struct
from flax import typing as flax_typing
from flax.metrics import tensorboard
import jax
import jax.numpy as jnp
import numpy as np
import optax


class TrainState(struct.PyTreeNode):
  """Train state."""

  train_model: nn.Module = struct.field(pytree_node=False)
  eval_model: nn.Module = struct.field(pytree_node=False)
  model_vars: Any = struct.field(pytree_node=True)
  tx: optax.GradientTransformation = struct.field(pytree_node=False)
  opt_state: optax.OptState = struct.field(pytree_node=True)


def create_train_state(
    rng: jax.Array,
    train_model: nn.Module,
    eval_model: nn.Module,
) -> TrainState:
  """Creates initial `TrainState`."""
  model_vars = train_model.init({'params': rng}, jnp.ones([1, 28, 28, 1]))
  learning_rate = 0.1
  momentum = 0.9
  tx = optax.sgd(learning_rate, momentum)
  return TrainState(
      train_model=train_model,
      eval_model=eval_model,
      model_vars=model_vars,
      tx=tx,
      opt_state=tx.init(model_vars['params']),
  )


def prepare_data_perm(
    ds: dict[str, jax.Array],
    batch_size: int,
    rng: jax.Array,
    num_steps: int = sys.maxsize,
) -> jax.Array:
  """Creates random permutation of data."""
  ds_size = len(ds['image'])
  num_steps = min(num_steps, ds_size // batch_size)
  perms = jax.random.permutation(rng, len(ds['image']))
  perms = perms[: num_steps * batch_size]  # skip incomplete batch
  return perms.reshape((num_steps, batch_size))


@functools.partial(jax.jit, static_argnums=(3,))
def _apply_model(
    model_vars: flax_typing.FrozenVariableDict,
    images: Any,
    labels: Any,
    apply_fn: Callable[..., Any],
) -> tuple[jax.Array, jax.Array, jax.Array, flax_typing.FrozenVariableDict]:
  """Computes gradients, loss and accuracy for a single batch."""

  def loss_fn(model_vars):
    logits, updated_vars = apply_fn(
        model_vars,
        images,
        rngs={'params': jax.random.PRNGKey(0)},
        mutable=True,
    )
    one_hot = jax.nn.one_hot(labels, 10)
    loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
    return loss, (logits, updated_vars)

  grad_fn = jax.value_and_grad(loss_fn, has_aux=True, allow_int=True)
  aux, grads = grad_fn(model_vars)
  loss, (logits, updated_vars) = aux
  accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)

  return grads, loss, accuracy, updated_vars


@jax.jit
def _update_model(state, grads, model_vars) -> TrainState:
  """Updates parameters in model_vars with given gradients."""
  params = state.model_vars['params']
  param_grad = grads['params']
  updates, new_opt_state = state.tx.update(param_grad, state.opt_state, params)
  new_params = optax.apply_updates(params, updates)
  model_vars.update(params=new_params)
  return state.replace(
      model_vars=model_vars,
      opt_state=new_opt_state,
  )


def _train_epoch(
    state: TrainState,
    train_ds: dict[str, jax.Array],
    batch_size: int,
    rng: jax.Array,
) -> tuple[TrainState, float, float]:
  """Trains for a single epoch."""
  perms = prepare_data_perm(train_ds, batch_size, rng)

  epoch_loss = []
  epoch_accuracy = []

  for perm in perms:
    batch_images = train_ds['image'][perm, ...]
    batch_labels = train_ds['label'][perm, ...]
    grads, loss, accuracy, updated_var = _apply_model(
        state.model_vars, batch_images, batch_labels, state.train_model.apply
    )
    state = _update_model(state, grads, updated_var)
    epoch_loss.append(loss)
    epoch_accuracy.append(accuracy)
  train_loss = np.mean(epoch_loss)
  train_accuracy = np.mean(epoch_accuracy)
  return state, train_loss, train_accuracy


def train_and_evaluate(
    num_epochs: int,
    workdir: str,
    train_ds: dict[str, jax.Array],
    test_ds: dict[str, jax.Array],
    state: TrainState,
) -> TrainState:
  """Execute model training and evaluation loop."""
  rng = jax.random.key(0)
  summary_writer = tensorboard.SummaryWriter(workdir)

  batch_size = 128
  for epoch in range(1, num_epochs + 1):
    rng, input_rng = jax.random.split(rng)
    state, train_loss, train_accuracy = _train_epoch(
        state, train_ds, batch_size, input_rng
    )
    _, test_loss, test_accuracy, _ = _apply_model(
        state.model_vars,
        test_ds['image'],
        test_ds['label'],
        state.eval_model.apply,
    )
    print(
        'epoch:% 3d, train_loss: %.30f, train_accuracy: %.30f, test_loss:'
        ' %.30f, test_accuracy: %.30f'
        % (
            epoch,
            train_loss,
            train_accuracy * 100,
            test_loss,
            test_accuracy * 100,
        ),
        flush=True,
    )
    summary_writer.scalar('train_loss', train_loss, epoch)
    summary_writer.scalar('train_accuracy', train_accuracy, epoch)
    summary_writer.scalar('test_loss', test_loss, epoch)
    summary_writer.scalar('test_accuracy', test_accuracy, epoch)

  summary_writer.flush()
  return state


def serve(
    serve_model: nn.Module,
    model_vars: flax_typing.FrozenVariableDict,
    test_ds: Any,
) -> jax.Array:
  """Serves a given model with given vars and computes test loss."""
  sample_image, sample_label = test_ds['image'][:64], test_ds['label'][:64]
  logits = jax.jit(serve_model.apply)(
      model_vars, sample_image, rngs={'params': jax.random.PRNGKey(0)}
  )
  # compute test loss
  one_hot = jax.nn.one_hot(sample_label, 10)
  loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
  return loss



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for transpose."""

import math
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import transpose
import jax.numpy as jnp


class AqtTransposeTest(parameterized.TestCase):

  @parameterized.parameters(
      ((10, 1, 10), (0, 2, 1)),
      ((10, 1, 5), (1, 0, 2)),
      ((5, 1, 10), (0, 1, 2)),
      ((10, 20, 30), (1, 2, 0)),
      ((10, 1, 1), (0, 2, 1)),
      ((10, 1, 1), (2, 1, 0)),
      ((10, 1, 1), (1, 2, 0)),
  )
  def test_transpose(self, tensor_shape, transpose_axes):
    t = jnp.arange(math.prod(tensor_shape)).reshape(tensor_shape)
    t_t = transpose.transpose(t, transpose_axes)
    self.assertTrue((t_t == jnp.transpose(t, transpose_axes)).all())

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 1, 4, 7, 1),
          (2, 6, 4, 7, 3),
          (2, 3, 4, 5),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 4, 1, 7, 1),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 1, 4, 1, 1, 1),
          (2, 6, 4, 7, 8, 3),
          (2, 3, 4, 5),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 4, 1, 1, 1, 1),
      ),
  )
  def test_lhs_scale_transpose_to_output(
      self,
      lhs_scale_shape,
      lhs_shape,
      rhs_shape,
      dimension_numbers,
      expected_qlhs_scale_t_shape
  ):
    lhs_scale = jnp.zeros(lhs_scale_shape)
    lhs = jnp.zeros(lhs_shape)
    rhs = jnp.zeros(rhs_shape)
    qlhs_scale_t = transpose.lhs_scale_transpose_to_output(
        lhs_scale, dimension_numbers, lhs.shape, rhs.shape
    )

    self.assertIsNotNone(qlhs_scale_t)
    self.assertEqual(qlhs_scale_t.shape, expected_qlhs_scale_t_shape)

    # Test recover.
    qlhs_scale_recovered = transpose.lhs_recover_scale_from_scale_t(
        qlhs_scale_t, dimension_numbers, lhs.shape, rhs_shape
    )
    self.assertEqual(lhs_scale.shape, qlhs_scale_recovered.shape)
    assert (lhs_scale == qlhs_scale_recovered).all()

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 3, 4, 5, 6),
          (2, 1, 4, 1),
          (2, 6, 4, 7),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 4, 1, 1, 1),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 3, 4, 5, 6, 7),
          (2, 1, 4, 1),
          (2, 7, 4, 8),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 4, 1, 1, 1, 1),
      ),
  )
  def test_rhs_scale_transpose_to_output(
      self,
      lhs_shape,
      rhs_scale_shape,
      rhs_shape,
      dimension_numbers,
      expected_qrhs_scale_t_shape
  ):
    lhs = jnp.zeros(lhs_shape)
    rhs = jnp.zeros(rhs_shape)
    rhs_scale = jnp.zeros(rhs_scale_shape)
    qrhs_scale_t = transpose.rhs_scale_transpose_to_output(
        rhs_scale, dimension_numbers, lhs.shape, rhs.shape
    )

    self.assertIsNotNone(qrhs_scale_t)
    self.assertEqual(qrhs_scale_t.shape, expected_qrhs_scale_t_shape)

    # Test recover.
    qrhs_scale_recovered = transpose.rhs_recover_scale_from_scale_t(
        qrhs_scale_t, dimension_numbers, lhs.shape, rhs_shape
    )
    self.assertEqual(rhs_scale.shape, qrhs_scale_recovered.shape)
    assert (rhs_scale == qrhs_scale_recovered).all()

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 1, 4, 1, 3),
          (2, 3, 4, 5),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 3, 4, 1),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 1, 4, 1, 1, 3),
          (2, 3, 4, 5),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 3, 4, 1),
      ),
  )
  def test_lhs_scale_transpose_for_rhs_input(
      self, lhs_scale_shape, rhs_shape, dimension_numbers, expected_shape
  ):
    lhs_scale = jnp.zeros(lhs_scale_shape)
    rhs = jnp.zeros(rhs_shape)
    result = transpose.lhs_scale_transpose_for_rhs_input(
        lhs_scale, dimension_numbers, rhs.shape
    )

    self.assertEqual(result.shape, expected_shape)

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 3, 4, 5, 6),
          (2, 6, 4, 1),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 1, 4, 1, 6),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 3, 4, 5, 6, 7),
          (2, 7, 4, 1),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 1, 4, 1, 1, 7),
      ),
  )
  def test_rhs_scale_transpose_for_lhs_input(
      self, lhs_shape, rhs_scale_shape, dimension_numbers, expected_shape
  ):
    lhs = jnp.zeros(lhs_shape)
    rhs_scale = jnp.zeros(rhs_scale_shape)
    result = transpose.rhs_scale_transpose_for_lhs_input(
        rhs_scale, dimension_numbers, lhs.shape
    )

    self.assertEqual(result.shape, expected_shape)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util functions to generate an experiment report after training.

Please refer to the README.md for an overview of the reporting tool.
"""

import dataclasses
import enum
import functools
import json
import os
import pathlib
import time
from typing import Any, Callable, Dict, List, Optional, Tuple

from aqt.jax_legacy.utils import hparams_utils
from aqt.jax_legacy.utils import tfevent_utils
import numpy as onp
from tensorflow.io import gfile

# Constants
COMPUTE_MEMORY_COST_FILENAME = 'compute_memory_cost.json'
HPARAMS_CONFIG_FILENAME = 'hparams_config.json'
REPORT_FILENAME = 'report.json'

EventSeries = tfevent_utils.EventSeries

# Type Aliases
# Nested dict mapping from component (first key), attribute (second key) to
# events stored in EventSeries. E.g. component = 'train', attribute = 'loss'.
_AllEvents = Dict[str, Dict[str, EventSeries]]

# Nested dict mapping from component (first key), attribute (second key) to
# aggregated metric (float). E.g. component = 'train', attribute = 'loss'.
_AllAggMetrics = Dict[str, Dict[str, float]]


@enum.unique
class MinOrMax(enum.Enum):
  """Aggregation function to use for finding early stopping step."""
  MIN = enum.auto()  # use min value for early stopping step.
  MAX = enum.auto()  # use max value for early stopping step.

  def get_func(self) -> Callable[[onp.ndarray], int]:
    """Returns function associated with enum option. See parent class."""
    if self == MinOrMax.MIN:
      return onp.nanargmin
    elif self == MinOrMax.MAX:
      return onp.nanargmax
    else:
      raise ValueError('MinOrMax enum option not recognized.')


@enum.unique
class SmoothingKernel(enum.Enum):
  """Kernel function to use for smoothing."""
  # RECTANGULAR:Every value in symmetric window weighted equally. Values
  # outside the window are not included in average.
  # TRIANGULAR: Every value in symmetric window weighted as a linear function of
  # absolute distance to kernel center. Values outside the window are not
  # included in average.
  RECTANGULAR = enum.auto()
  TRIANGULAR = enum.auto()

  def rectangular_kernel(self, x: float, window_size_in_steps: int) -> float:
    """Rectangular kernel for moving window average.

    All values in window are equally weighted.

    Args:
      x: Distance to kernel center in steps.
      window_size_in_steps: Size of the window to average over.

    Returns:
      Unnormalized weight to use for averaging, e.g. in `np.average()`.

    Raises:
      ValueError: If window_size_in_steps arg is less than 1.
    """
    if window_size_in_steps < 1:
      raise ValueError('window_size_in_steps has to be >= 1.')
    if abs(x) <= window_size_in_steps / 2:
      return 1.0
    else:
      return 0.0

  def triangular_kernel(self, x: float, window_size_in_steps: int) -> float:
    """Triangular kernel for moving window average.

    The weight is a linear function of the absolute distance to the kernel
    center.

    Args:
      x: Distance to kernel center in steps.
      window_size_in_steps: Size of the window to average over.

    Returns:
      Unnormalized weight to use for averaging, e.g. in `np.average()`.

    Raises:
      ValueError: If window_size_in_steps arg is less than 1.
    """
    if window_size_in_steps < 1:
      raise ValueError('window_size_in_steps has to be >= 1.')
    return max(0.0, window_size_in_steps / 2 - abs(x))

  def get_func(
      self,
      window_size_in_steps: Optional[int] = None) -> Callable[[float], float]:
    """Returns function associated with enum option. See parent class."""
    if self == SmoothingKernel.RECTANGULAR:
      if window_size_in_steps is None:
        raise ValueError('For rectangular smoothing_kernel '
                         'window_size_in_steps must be provided.')
      return functools.partial(
          self.rectangular_kernel, window_size_in_steps=window_size_in_steps)
    elif self == SmoothingKernel.TRIANGULAR:
      if window_size_in_steps is None:
        raise ValueError('For triangular smoothing_kernel '
                         'window_size_in_steps must be provided.')
      return functools.partial(
          self.triangular_kernel, window_size_in_steps=window_size_in_steps)
    else:
      raise ValueError('SmoothingKernel enum option not recognized.')




@dataclasses.dataclass
class ExperimentReport:
  """Report for a single experiment run based on its TFEvents files."""
  # Model directory corresponding to single run, with TFEvents files to
  # generate report from.
  model_dir: str

  # Metrics at early stop step, with smoothing applied.
  # If NaN values present, then this field will
  # be left None, but unsmoothed_metrics will still be reported.
  # maps component name (e.g. eval) to metrics dict, which in turn maps
  # attribute name to scalar value.
  metrics: Optional[_AllAggMetrics]

  # Metrics without smoothing at early stop step.
  # maps component name (e.g. eval) to metrics dict, which in turn maps
  # attribute name to scalar value.
  unsmoothed_metrics: Optional[_AllAggMetrics]

  # Step at which early_stop_attr in early_stop_ds_dir is minimized. Scalars are
  # reported at this step.
  early_stop_step: int

  # Number of training steps. In combination with early_stop_step, can help
  # determine whether training converged and started to overfit.
  num_train_steps: int

  # Arguments passed into create_end_of_training_report(), the function that
  # created this report.
  # Included here because the arguments can impact the reported metrics, e.g.
  # which attribute was used to find the early stopping step.
  report_query_args: Dict[str, Any]

  # Human-readable experiment name.
  experiment_name: Optional[str] = None

  # Name of user who launched the experiment.
  user_name: Optional[str] = None

  # When experiment was launched, formatted as '%Y%m%dT%H%M%S'.
  launch_time: Optional[str] = None

  # Evaluation frequency. How often summaries were saved to file.
  eval_freq: Optional[int] = None

  # If any metrics contain NaN values, first step at which a NaN value occurs.
  first_nan_step: Optional[int] = None

  # Tensorboard ID or URL.
  tensorboard_id: Optional[str] = None

  # Full path to hparams config file.
  hparams_config_path: Optional[str] = None



def check_for_nans(event_series: EventSeries, start_step: int) -> Optional[int]:
  """Finds step >= start_step at which first NaN value occurs if there are any.

  Args:
    event_series: list of tuples (step, value).
    start_step: After which step to check for NaNs.

  Returns:
    Step at which first NaN value occurs, or None otherwise.

  """
  keep_indices = (event_series.steps >= start_step)
  event_series.steps = event_series.steps[keep_indices]
  event_series.values = event_series.values[keep_indices]
  nan_indices = onp.argwhere(onp.isnan(event_series.values))
  if nan_indices.size:
    return int(event_series.steps[onp.min(nan_indices)])
  return None


def check_all_events_for_nans(all_events: _AllEvents) -> Optional[int]:
  """Finds step at which first NaN value occurs if there are any.

  Args:
    all_events: Nested dict mapping from component, attribute, to EventSeries.

  Returns:
    Step at which first NaN value occurs, or None otherwise.
  """

  first_nan_step = None

  for events_dict in all_events.values():
    for events in events_dict.values():
      cur_first_nan_step = check_for_nans(events, start_step=0)
      if cur_first_nan_step is None:
        continue
      if first_nan_step is None:
        first_nan_step = cur_first_nan_step
      else:
        first_nan_step = min(first_nan_step, cur_first_nan_step)
  return first_nan_step


def find_early_stop_step(event_series: EventSeries,
                         early_stop_func: Callable[[onp.ndarray], int],
                         start_step: int) -> int:
  """Finds step >= start_step at which event_series is minimized.

  Args:
    event_series: list of tuples (step, value).
    early_stop_func: Aggregator function to use to find early_stop_step.
    start_step: After which step to include values in moving average.

  Returns:
    Step at which moving average of series is minimized.

  """
  keep_indices = (event_series.steps >= start_step)
  event_series.steps = event_series.steps[keep_indices]
  event_series.values = event_series.values[keep_indices]

  if event_series.steps.size == 0:
    raise ValueError('event_series does not have events after start_step.')

  if onp.all(onp.isnan(event_series.values)):
    return start_step
  early_stop_idx = early_stop_func(event_series.values)
  return int(event_series.steps[early_stop_idx])


def apply_smoothing_about_step(events: EventSeries, step: int,
                               kernel_fn: Callable[[float], float]) -> float:
  """Applies smoothing of event values for a single step.

  Args:
    events: list of tuples (step, value).
    step: Step to apply smoothing about.
    kernel_fn: Kernel function to use for smoothing.

  Returns:
    Smoothed value at step.

  Raises:
    ValueError: If NaN values present in events.values.
  """
  if check_for_nans(events, start_step=0) is not None:
    raise ValueError(
        'NaN values encountered in smoothing, which is not supported.')
  weights = onp.vectorize(kernel_fn)(events.steps - step)
  return float(onp.average(events.values, weights=weights))


def apply_smoothing(events: EventSeries,
                    kernel_fn: Callable[[float], float]) -> EventSeries:
  """Applies smoothing of event values over all steps.

  Args:
    events: list of tuples (step, value).
    kernel_fn: Kernel function to use for smoothing.

  Returns:
    Smoothed events for all steps in steps arg.
  """
  smoothed_events = EventSeries(
      name=events.name,
      steps=onp.array([], dtype=int),
      values=onp.array([]),
      wall_times=None)
  for i in range(len(events.steps)):
    smoothed_events.steps = onp.append(smoothed_events.steps, events.steps[i])
    smoothed_value = apply_smoothing_about_step(
        events=events, step=events.steps[i], kernel_fn=kernel_fn)
    smoothed_events.values = onp.append(smoothed_events.values, smoothed_value)

  return smoothed_events


def get_agg_metrics_at_step(
    all_events: _AllEvents, step: int,
    smoothing_kernel_fn: Optional[Callable[[float], float]]) -> _AllAggMetrics:
  """Computes aggregated metrics from EventSeries dicts at early stop step.

  Args:
    all_events: Nested dict mapping from component, attribute, to EventSeries.
    step: Step at which to get event values to compute aggregated metrics.
    smoothing_kernel_fn: If None, no smoothing will be applied. If any NaNs are
      present, has to be set to None, otherwise ValueError will be raised.

  Returns:
    dict mapping from (component, attribute) to aggregated scalar metric.

  """
  all_agg_metrics = {}
  for component, events_dict in all_events.items():
    agg_metrics_dict = {}
    for attr, events in events_dict.items():
      if smoothing_kernel_fn is None:
        index = onp.argmin(onp.abs(events.steps - step))
        agg_metrics_dict[attr] = events.values[index]
      else:
        agg_metrics_dict[attr] = apply_smoothing_about_step(
            events, step=step, kernel_fn=smoothing_kernel_fn)
    all_agg_metrics[str(component)] = agg_metrics_dict

  return all_agg_metrics


def compute_agg_metrics_from_events(
    all_events: _AllEvents,
    early_stop_component: str,
    early_stop_attr: str,
    early_stop_agg: MinOrMax,
    smoothing_kernel: SmoothingKernel,
    window_size_in_steps: Optional[int] = None,
    start_step: int = 0
) -> Tuple[_AllAggMetrics, Optional[_AllAggMetrics], int, Optional[int]]:
  """Computes aggregated metrics from EventSeries dicts.

  Args:
    all_events: Nested dict mapping from component, attribute, to EventSeries.
    early_stop_component: Which component to use to find early_stop_step.
    early_stop_attr: Attribute to find minimum or maximum of, e.g. 'perplexity'.
    early_stop_agg: Which aggregator to use to find early_stop_step. See
      MinOrMax class for enum options.
    smoothing_kernel: Which kernel to use for smoothing. See SmoothingKernel
      class for enum options.
    window_size_in_steps: Only applicable to some kernels, including
      'rectangular' kernel. Number of steps to average over.
    start_step: After which step to consider early stopping, e.g. if set to 100,
      only steps >= 100 will be considered.

  Returns:
    Tuple of dict mapping from (component, attribute) to aggregated scalar
      metric and early_stop_step.

  """
  first_nan_step = check_all_events_for_nans(all_events=all_events)

  early_stop_func = early_stop_agg.get_func()
  early_stop_events = all_events[early_stop_component][early_stop_attr]

  if first_nan_step is None:
    # Apply smoothing to early stop component events.
    smoothing_kernel_func = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    early_stop_events = apply_smoothing(
        events=early_stop_events, kernel_fn=smoothing_kernel_func)

  early_stop_step = find_early_stop_step(
      early_stop_events, early_stop_func=early_stop_func, start_step=start_step)

  all_agg_metrics_unsmoothed = get_agg_metrics_at_step(
      all_events=all_events, step=early_stop_step, smoothing_kernel_fn=None)

  if first_nan_step is None:
    # Only get smoothed metrics if no NaN values found.
    all_metrics_smoothed = get_agg_metrics_at_step(
        all_events=all_events,
        step=early_stop_step,
        smoothing_kernel_fn=smoothing_kernel_func)
  else:
    all_metrics_smoothed = None

  return all_agg_metrics_unsmoothed, all_metrics_smoothed, early_stop_step, first_nan_step


def create_end_of_training_report(
    model_dir: str,
    eval_freq: int,
    num_train_steps: int,
    early_stop_attr: str,
    early_stop_agg: MinOrMax,
    smoothing_kernel: SmoothingKernel,
    early_stop_ds_dir: Optional[str] = None,
    other_ds_dirs: Optional[List[str]] = None,
    tags_to_include: Optional[List[str]] = None,
    window_size_in_steps: int = 1,
    start_step: int = 0,
    experiment_name: Optional[str] = None,
    user_name: Optional[str] = None,
    launch_time: Optional[str] = None,
    tensorboard_id: Optional[str] = None,
) -> ExperimentReport:
  """Creates an experiment report from TFEvents data after training completion.

  Args:
    model_dir: A model directory corresponding to a single model run, with
      TFEvent file(s) and a single hparams_config file. The TFEvent files can
      either be stored directly in model_dir, or in subdirectories in model_dir,
      but not both.
    eval_freq: Frequency of event saving.
    num_train_steps: Number of training steps.
    early_stop_attr: Attribute to find minimum or maximum of, e.g. 'perplexity'.
    early_stop_agg: Which aggregator to use to find early_stop_step. See
      MinOrMax class for enum options.
    smoothing_kernel: Which kernel to use for smoothing. See SmoothingKernel
      class for enum options.
    early_stop_ds_dir: The events subdir in model_dir to use to find
      early_stop_step if model_dir has subdirs. The early_stop_attr within
      early_stop_ds_dir will be used to find the early_stop_step.
    other_ds_dirs: List of other subdirs in model_dir with events to report.
    tags_to_include: List of event tags that should be included.
    window_size_in_steps: Number of steps to average over. Should be multiple of
      eval_freq. If set to 1, no averaging will be applied.
    start_step: After which step to consider early stopping, e.g. if set to 100,
      only steps >= 100 will be considered.
    experiment_name:  Human-readable experiment name.
    user_name: Name of user who launched the experiment.
    launch_time: When experiment was launched, formatted as '%Y%m%dT%H%M%S'.
    tensorboard_id: Tensorboard ID, e.g. URL to tensorboard dev, if applicable.

  Returns:
    An ExperimentReport dataclass instance.

  """
  # Saving report query args, to be included in the report.
  report_query_args = {
      'early_stop_attr': early_stop_attr,
      'early_stop_agg': early_stop_agg.name,
      'early_stop_ds_dir': early_stop_ds_dir,
      'other_ds_dirs': other_ds_dirs,
      'tags_to_include': tags_to_include,
      'smoothing_kernel': smoothing_kernel.name,
      'window_size_in_steps': window_size_in_steps,
      'start_step': start_step,
  }

  all_events = {}

  early_stop_component = None

  # If subdirs provided
  if early_stop_ds_dir is not None or other_ds_dirs is not None:
    if early_stop_ds_dir is None:
      raise ValueError(
          'If other_ds_dirs is not None, early_stop_ds_dir has to be '
          'provided.')

    early_stop_events = tfevent_utils.get_parsed_tfevents(
        os.path.join(model_dir, early_stop_ds_dir), tags_to_include)
    early_stop_component = early_stop_ds_dir
    all_events[early_stop_component] = early_stop_events

    if other_ds_dirs is not None:
      for ds_dir in other_ds_dirs:
        if ds_dir is not None:
          all_events[ds_dir] = tfevent_utils.get_parsed_tfevents(
              os.path.join(model_dir, ds_dir), tags_to_include)
  else:
    # If no subdirs provided, will assume that there are no subcomponents.
    # For consistency with the case when we do have components, we store the
    # events under the dummy component 'all'.
    early_stop_component = 'all'
    all_events[early_stop_component] = tfevent_utils.get_parsed_tfevents(
        model_dir, tags_to_include)

  all_agg_metrics_unsmoothed, all_agg_metrics_smoothed, early_stop_step, first_nan_step = compute_agg_metrics_from_events(
      all_events=all_events,
      early_stop_component=early_stop_component,
      early_stop_attr=early_stop_attr,
      early_stop_agg=early_stop_agg,
      smoothing_kernel=smoothing_kernel,
      window_size_in_steps=window_size_in_steps,
      start_step=start_step)

  report = ExperimentReport(
      model_dir=model_dir,
      metrics=all_agg_metrics_smoothed,
      unsmoothed_metrics=all_agg_metrics_unsmoothed,
      early_stop_step=early_stop_step,
      num_train_steps=num_train_steps,
      eval_freq=eval_freq,
      first_nan_step=first_nan_step,
      experiment_name=experiment_name,
      user_name=user_name,
      launch_time=launch_time,
      tensorboard_id=tensorboard_id,
      report_query_args=report_query_args,
  )

  return report


def report_path_from_model_dir(model_dir: pathlib.Path) -> pathlib.Path:
  """Generates report path from model_dir."""
  report_path = '--'.join(list(model_dir.parts)[1:])

  # get rid of file extensions if there were any
  report_path = os.path.splitext(str(report_path))[0]

  return pathlib.Path(report_path)


def save_report(report: ExperimentReport,
                report_dir: pathlib.Path) -> pathlib.Path:
  """Saves json serialized ExperimentReport instance to report directory.

  If report has hparams_config_path, will attempt to copy over the config file
  to the report_dir as well.

  The filename is auto-generated by concatenating report.model_dir.

  Args:
    report: The report to be saved.
    report_dir: The directory to save to.

  Returns:
    Path where report was saved to.
  """
  report_path = report_path_from_model_dir(pathlib.Path(report.model_dir))
  if not gfile.exists(report_dir / report_path):
    gfile.makedirs(report_dir / report_path)
  if report.hparams_config_path is not None:
    gfile.copy(
        report.hparams_config_path,
        report_dir / report_path / HPARAMS_CONFIG_FILENAME,
        overwrite=True)
    report.hparams_config_path = str(report_dir / report_path /
                                     HPARAMS_CONFIG_FILENAME)
  data_dict = dataclasses.asdict(report)
  path = pathlib.Path(report_dir / report_path / REPORT_FILENAME)
  with gfile.GFile(path, 'w') as file:
    json.dump(data_dict, file, indent=2)
  return report_dir / report_path





# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for flax e2e model."""

import copy
import functools

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config
from aqt.jax.v2 import utils
from aqt.jax.v2.examples import flax_e2e_model
from aqt.jax.v2.flax import aqt_flax_calibration
from aqt.jax.v2.flax import delayed_scaling_calibration
import jax
import jax.numpy as jnp
import numpy as np


def _dummy_dataset(ds_size, image_rng, label_rng):
  return {
      "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
      "label": jax.random.randint(
          key=label_rng, shape=(ds_size,), minval=0, maxval=10
      ),
  }


class MnistTest(parameterized.TestCase):

  # Unable to use config_v4() in parameters since it needs jax.device info.
  # TODO(aqt): Move config_v4() into parameters once int4 works for cpu.
  @parameterized.parameters([
      (
          {
              "drhs_bits": 8,
              "drhs_accumulator_dtype": jnp.int32,  # overwrite the default None
          },
          8,
      ),
      (
          {
              "fwd_bits": 4,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          4,
      ),
  ])
  def test_mnist_training(self, configs, bits):
    aqt_cfg_dg = config.config_v4(**configs)
    target_loss = {
        8: {
            "cpu": [
                3.122317314147949218750000000000,
                3.122316360473632812500000000000,
                3.122316837310791015625000000000,  # colab
            ],
            "TPU v2": [3.198328018188476562500000000000],
            "TPU v3": [3.198328018188476562500000000000],
            "TPU v4": [3.198297500610351562500000000000],
            "TPU v5 lite": [3.198297500610351562500000000000],
        },
        4: {
            "cpu": [2.258865118026733398437500000000],
            "TPU v2": [2.302409172058105468750000000000],
            "TPU v3": [2.302409172058105468750000000000],
            "TPU v4": [2.302409172058105468750000000000],
            "TPU v5 lite": [2.302409172058105468750000000000],
        },
    }
    # below 3 lines are differences between config_v4/v3 and fully_quantized
    config.set_stochastic_rounding(aqt_cfg_dg, True, True, "jax.uniform")
    aqt_cfg_dg.dlhs.rhs.use_fwd_quant = True
    aqt_cfg_dg.drhs.rhs.use_fwd_quant = True

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)

    # Stage 1: regular training
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, train_loss, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size=ds_size // 2, rng=input_rng
    )

    device_kind = jax.devices()[0].device_kind
    expected_train_loss = target_loss[bits][device_kind]
    if train_loss not in expected_train_loss:
      msg = "train_loss changed. Consider updating with the following:\n"
      msg += f'        "{device_kind}": [{train_loss:.30f}]'
      self.fail(msg)

    # Run forward once more in the same mode to get logits for testing below.
    logits_s1, _ = forward(state.model, state.cnn_eval.apply)

    # Stage 2: Model conversion (quantized weights freezing)

    apply_serving, model_serving = flax_e2e_model.serving_conversion(state)

    dtype = jnp.dtype
    expected_dtype = jnp.int4 if bits == 4 else jnp.int8
    expected_aqt_pytree = {
        "aqt": {
            "AqtEinsum_0": {
                "AqtDotGeneral_0": {
                    "qlhs": {
                        "frozen": aqt_tensor.QTensor(
                            qvalue=(expected_dtype, (2, 5, 10)),
                            scale=[(dtype("float32"), (2, 1, 10))],
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                        )
                    }
                }
            },
            "Dense_0": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        "frozen": aqt_tensor.QTensor(
                            # The weight shape was (3136, 256) before tiling.
                            # After tiling it is (2, 1568, 256).
                            # Contraction shape 3136 is tiled to (2, 1568).
                            # The remaining shape 256 is not tiled.
                            qvalue=(expected_dtype, (2, 1568, 256)),
                            scale=[(dtype("float32"), (2, 1, 256))],
                            # The scale_t shape was (1, 256) before tiling.
                            # After tiling the scale shape is (2, 1, 256),
                            # then transposed to (2, 1, 256).
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                        )
                    }
                }
            },
            "Dense_1": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        "frozen": aqt_tensor.QTensor(
                            # The weight shape was (256, 10) before tiling.
                            # After tiling it is (2, 128, 10).
                            # Contraction shape 256 is tiled to (2, 128).
                            # The remaining shape 10 is not tiled.
                            qvalue=(expected_dtype, (2, 128, 10)),
                            scale=[(dtype("float32"), (2, 1, 10))],
                            # The scale_t shape was (1, 10) before tiling.
                            # After tiling the scale shape is (2, 1, 10),
                            # then transposed to (2, 1, 10).
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                        )
                    }
                }
            },
        },
        "batch_stats": {
            "BatchNorm_0": {
                "mean": (dtype("float32"), (32,)),
                "var": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "mean": (dtype("float32"), (64,)),
                "var": (dtype("float32"), (64,)),
            },
        },
        "params": {
            "BatchNorm_0": {
                "bias": (dtype("float32"), (32,)),
                "scale": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "bias": (dtype("float32"), (64,)),
                "scale": (dtype("float32"), (64,)),
            },
            "Conv_0": {
                "bias": (dtype("float32"), (32,)),
                "kernel": (dtype("float32"), (3, 3, 1, 32)),
            },
            "Conv_1": {
                "bias": (dtype("float32"), (64,)),
                "kernel": (dtype("float32"), (3, 3, 32, 64)),
            },
            "Dense_0": {
                "bias": (dtype("float32"), (256,)),
                "kernel": (dtype("float32"), (3136, 256)),
            },
            "Dense_1": {
                "bias": (dtype("float32"), (10,)),
                "kernel": (dtype("float32"), (256, 10)),
            },
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    if serving_pytree != expected_aqt_pytree:
      print()
      print("serving_pytree:      ", serving_pytree)
      print("expected_aqt_pytree: ", expected_aqt_pytree)
      assert False, "serving_pytree != expected_aqt_pytree"

    def zero_out_params(model, layer: str):
      updated_model = copy.deepcopy(model)
      updated_model["params"][layer]["kernel"] = jnp.zeros_like(
          updated_model["params"][layer]["kernel"]
      )
      return updated_model

    # We can, but do not have to delete unquantized weights.
    model_serving = zero_out_params(model_serving, "Dense_0")
    model_serving = zero_out_params(model_serving, "Dense_1")

    # Stage 3: inference mode.
    logits_s3, _ = forward(model_serving, apply_serving)
    assert (logits_s3 == logits_s1).all()

    # Sanity check 1: We can't zero out Conv_0 because it was not frozen.
    sanity_model = zero_out_params(model_serving, "Conv_0")
    bad_logits, _ = forward(sanity_model, apply_serving)
    assert not (bad_logits == logits_s3).all()

    # Sanity check 2: Frozen weights are indeed used for inference.
    #   If we zero them out, loss would change.
    qt = model_serving["aqt"]["Dense_0"]["AqtDotGeneral_0"]["qrhs"]["frozen"]
    qt.qvalue = jnp.zeros(qt.qvalue.shape).astype(qt.qvalue.dtype)
    bad_logits, _ = forward(model_serving, apply_serving)
    assert not (bad_logits == logits_s3).all()

  @parameterized.parameters([
      (
          {
              "fwd_bits": 8,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          8,
      ),
      (
          {
              "fwd_bits": 4,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          4,
      ),
  ])
  def test_mnist_training_quantized_conv(self, configs, bits):
    aqt_cfg_dg = config.config_v4(**configs)
    aqt_cfg_conv = config.conv_general_dilated_make(
        2, lhs_bits=bits, rhs_bits=bits, initialize_calibration=False
    )
    if aqt_cfg_conv.lhs:
      aqt_cfg_conv.dg_quantizer.lhs.init_calibration()
    if aqt_cfg_conv.rhs:
      aqt_cfg_conv.dg_quantizer.rhs.init_calibration()
    device_kind = jax.devices()[0].device_kind
    if device_kind == "cpu" and bits == 4:
      # Some 4-bit operations are not supported on cpu.
      # Omitting tests on cpu with 4-bits.
      return
    target_loss = {
        8: {
            "cpu": [2.302584648132324218750000000000],
            "TPU v2": [2.302581071853637695312500000000],
            "TPU v3": [2.302581071853637695312500000000],
            "TPU v4": [2.302581071853637695312500000000,],
            "TPU v5 lite": [2.302581071853637695312500000000],
        },
        4: {
            "TPU v2": [2.302409172058105468750000000000],
            "TPU v3": [2.302409172058105468750000000000],
            "TPU v4": [2.302409172058105468750000000000],
            "TPU v5 lite": [2.302409172058105468750000000000],
        },
    }
    # below 3 lines are differences between config_v4/v3 and fully_quantized
    config.set_stochastic_rounding(aqt_cfg_dg, True, True, "jax.uniform")
    aqt_cfg_dg.dlhs.rhs.use_fwd_quant = True
    aqt_cfg_dg.drhs.rhs.use_fwd_quant = True

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)

    # Stage 1: regular training
    state = flax_e2e_model.create_train_state(
        init_rng, aqt_cfg_dg=aqt_cfg_dg, aqt_cfg_conv=aqt_cfg_conv)

    state, train_loss, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size=ds_size // 2, rng=input_rng
    )

    device_kind = jax.devices()[0].device_kind
    expected_train_loss = target_loss[bits][device_kind]
    if train_loss not in expected_train_loss:
      msg = "train_loss changed. Consider updating with the following:\n"
      msg += f'        "{device_kind}": [{train_loss:.30f}]'
      self.fail(msg)

    # Run forward once more in the same mode to get logits for testing below.
    logits_s1, _ = forward(state.model, state.cnn_eval.apply)

    # Stage 2: Model conversion (quantized weights freezing)

    apply_serving, model_serving = flax_e2e_model.serving_conversion(state)

    dtype = jnp.dtype
    expected_dtype = jnp.int4 if bits == 4 else jnp.int8
    expected_aqt_pytree = {
        "aqt": {
            "AqtEinsum_0": {
                "AqtDotGeneral_0": {
                    "qlhs": {
                        "frozen": aqt_tensor.QTensor(
                            qvalue=(expected_dtype, (2, 5, 10)),
                            scale=[(dtype("float32"), (2, 1, 10))],
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                            tiling_state=None,
                        )
                    }
                }
            },
            "Conv_0": {
                "AqtConvGeneralDilated_0": {
                    "qrhs": {
                        "frozen": aqt_tensor.QTensor(
                            qvalue=(expected_dtype, (3, 3, 1, 32)),
                            scale=[(dtype("float32"), (1, 1, 1, 32))],
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                            tiling_state=None)
                    }
                }
            },
            "Conv_1": {
                "AqtConvGeneralDilated_0": {
                    "qrhs": {"frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (3, 3, 32, 64)),
                        scale=[(dtype("float32"), (1, 1, 1, 64))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                        tiling_state=None,
                        )
                             }
                    }
                },
            "Dense_0": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        "frozen": aqt_tensor.QTensor(
                            # The weight shape was (3136, 256) before tiling.
                            # After tiling it is (2, 1568, 256).
                            # Contraction shape 3136 is tiled to (2, 1568).
                            # The remaining shape 256 is not tiled.
                            qvalue=(expected_dtype, (2, 1568, 256)),
                            scale=[(dtype("float32"), (2, 1, 256))],
                            # The scale_t shape was (1, 256) before tiling.
                            # After tiling the scale shape is (2, 1, 256),
                            # then transposed to (2, 1, 256).
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                            tiling_state=None,
                        )
                    }
                }
            },
            "Dense_1": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        "frozen": aqt_tensor.QTensor(
                            # The weight shape was (256, 10) before tiling.
                            # After tiling it is (2, 128, 10).
                            # Contraction shape 256 is tiled to (2, 128).
                            # The remaining shape 10 is not tiled.
                            qvalue=(expected_dtype, (2, 128, 10)),
                            scale=[(dtype("float32"), (2, 1, 10))],
                            # The scale_t shape was (1, 10) before tiling.
                            # After tiling the scale shape is (2, 1, 10),
                            # then transposed to (2, 1, 10).
                            scale_t=None,
                            bias=[],
                            dequant_dtype=dtype("float32"),
                            tiling_state=None,
                        )
                    }
                }
            },
        },
        "batch_stats": {
            "BatchNorm_0": {
                "mean": (dtype("float32"), (32,)),
                "var": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "mean": (dtype("float32"), (64,)),
                "var": (dtype("float32"), (64,)),
            },
        },
        "params": {
            "BatchNorm_0": {
                "bias": (dtype("float32"), (32,)),
                "scale": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "bias": (dtype("float32"), (64,)),
                "scale": (dtype("float32"), (64,)),
            },
            "Conv_0": {
                "bias": (dtype("float32"), (32,)),
                "kernel": (dtype("float32"), (3, 3, 1, 32)),
            },
            "Conv_1": {
                "bias": (dtype("float32"), (64,)),
                "kernel": (dtype("float32"), (3, 3, 32, 64)),
            },
            "Dense_0": {
                "bias": (dtype("float32"), (256,)),
                "kernel": (dtype("float32"), (3136, 256)),
            },
            "Dense_1": {
                "bias": (dtype("float32"), (10,)),
                "kernel": (dtype("float32"), (256, 10)),
            },
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )

    if serving_pytree != expected_aqt_pytree:
      print()
      print("serving_pytree:      ", serving_pytree)
      print("expected_aqt_pytree: ", expected_aqt_pytree)
      assert False, "serving_pytree != expected_aqt_pytree"

    def zero_out_params(model, layer: str):
      updated_model = copy.deepcopy(model)
      updated_model["params"][layer]["kernel"] = jnp.zeros_like(
          updated_model["params"][layer]["kernel"]
      )
      return updated_model

    # We can, but do not have to delete unquantized weights.
    model_serving = zero_out_params(model_serving, "Dense_0")
    model_serving = zero_out_params(model_serving, "Dense_1")
    model_serving = zero_out_params(model_serving, "Conv_0")
    model_serving = zero_out_params(model_serving, "Conv_1")

    # Stage 3: inference mode.
    logits_s3, _ = forward(model_serving, apply_serving)
    assert (logits_s3 == logits_s1).all()

    # Sanity check 2: Frozen weights are indeed used for inference.
    #   If we zero them out, loss would change.
    qt = model_serving[
        "aqt"
        ]["Conv_0"]["AqtConvGeneralDilated_0"]["qrhs"]["frozen"]
    qt.qvalue = jnp.zeros(qt.qvalue.shape).astype(qt.qvalue.dtype)
    bad_logits, _ = forward(model_serving, apply_serving)
    assert not (bad_logits == logits_s3).all()

  @parameterized.parameters([
      (
          {
              "drhs_bits": 8,
              "drhs_accumulator_dtype": jnp.int32,  # overwrite the default None
          },
          8,
      ),
      (
          {
              "fwd_bits": 4,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          4,
      ),
  ])
  def test_mnist_calibration(self, configs, bits):
    aqt_cfg_dg = config.config_v4(**configs)
    device_kind = jax.devices()[0].device_kind
    if device_kind == "cpu" and bits == 4:
      # Some 4-bit operations are not supported on cpu.
      # Omitting tests on cpu with 4-bits.
      return

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng1, image_rng2 = jax.random.split(rng, 3)
    rng, label_rng1, label_rng2 = jax.random.split(rng, 3)
    rng, input_rng = jax.random.split(rng)
    rng, calibration_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 64
    batch_size = 8
    ds = _dummy_dataset(ds_size, image_rng1, label_rng1)
    ds2 = _dummy_dataset(ds_size, image_rng2, label_rng2)

    # Stage 1: regular training
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, _, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size, rng=input_rng
    )

    # Stage 2: Calibration.
    flax_e2e_model.update_cfg_with_calibration(state.cnn_train.aqt_cfg_dg)
    flax_e2e_model.update_cfg_with_calibration(state.cnn_eval.aqt_cfg_dg)
    calibrate_f, model_calibrate = flax_e2e_model.calibration_conversion(state)

    calibration_steps = 4
    calibrated_params = flax_e2e_model.calibrate_epoch(
        calibrate_f,
        model_calibrate,
        ds,
        batch_size,
        rng=calibration_rng,
        calibration_steps=calibration_steps,
    )
    calibration_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), calibrated_params
    )
    dtype = jnp.dtype
    expected_calibration_pytree = {
        "AqtEinsum_0": {
            # For the Einsum case, lhs and rhs are swapped.
            "AqtDotGeneral_0": {
                "MeanOfAbsMaxCalibration_0": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (1, 1, 1)),
                },
                "MeanOfAbsMaxCalibration_1": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (2, 1, 10)),
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "MeanOfAbsMaxCalibration_0": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (1, 1, 1)),
                },
                "MeanOfAbsMaxCalibration_1": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (2, 1, 256)),
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "MeanOfAbsMaxCalibration_0": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (1, 1, 1)),
                },
                "MeanOfAbsMaxCalibration_1": {
                    "count": (dtype("int32"), ()),
                    "sum_of_max": (dtype("float32"), (2, 1, 10)),
                },
            }
        },
    }
    utils.test_pprint_eq(expected_calibration_pytree, calibration_pytree["qc"])

    # The count number should be equal to the number of calibration.
    einsum_params = calibrated_params["qc"]["AqtEinsum_0"]["AqtDotGeneral_0"]
    einsum_count = einsum_params["MeanOfAbsMaxCalibration_0"]["count"]
    self.assertEqual(calibration_steps, einsum_count)

    # Stage 3: Training with the calibrated numbers.
    state = state.replace(model=copy.deepcopy(calibrated_params))
    state, _, _ = flax_e2e_model.train_epoch(
        state, ds2, batch_size, rng=input_rng
    )

    # The calibrated parameters must not change.
    jax.tree.map(
        np.testing.assert_array_equal,
        calibrated_params["qc"],
        state.model["qc"],
    )

    # Other parameters should change due to the training.
    def assert_array_not_equal(x, y):
      mean_err = jnp.mean(jnp.abs(x - y))
      if mean_err == 0.0:
        assert False

    jax.tree.map(
        assert_array_not_equal,
        calibrated_params["params"],
        state.model["params"],
    )

    # Stage 4. Convert the calibrated checkpoint.
    serve_fn, model_serving = flax_e2e_model.serving_conversion(
        state, weight_only=False
    )
    dtype = jnp.dtype
    expected_dtype = dtype("int4") if bits == 4 else dtype("int8")
    expected_aqt_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 5, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 1568, 256)),
                        scale=[(dtype("float32"), (2, 1, 256))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 128, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    utils.test_pprint_eq(expected_aqt_pytree, serving_pytree["aqt"])

    # Compare logits of models before conversion and after conversion.
    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    logits_before_conversion, _ = forward(state.model, state.cnn_eval.apply)
    logits_after_conversion, _ = forward(model_serving, serve_fn)
    assert (logits_before_conversion == logits_after_conversion).all()

  @parameterized.parameters([
      (
          {
              "drhs_bits": 8,
          },
          8,
      ),
      (
          {
              "fwd_bits": 4,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          4,
      ),
  ])
  def test_mnist_weighted_stats_calibration(self, configs, bits):
    aqt_cfg_dg = config.config_v4(**configs)
    device_kind = jax.devices()[0].device_kind
    if device_kind == "cpu" and bits == 4:
      # Some 4-bit operations are not supported on cpu.
      # Omitting tests on cpu with 4-bits.
      return

    # Update cfg with WeightedStatsCalibration.
    calibration_cls = functools.partial(
        aqt_flax_calibration.WeightedStatsCalibration,
        l1_dev_coeff=1.0,
        lp_dev_coeff=1.0,
        max_dev_coeff=1.0,
        const_bound_coeff=1.0,
        quant_collection="qc",
    )
    config.set_fwd_calibration(aqt_cfg_dg, calibration_cls)
    aqt_cfg_dg.fwd.dg_quantizer.lhs.calib_shared_axes = "per_tensor"

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng, label_rng = jax.random.split(rng, 3)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 64
    batch_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)

    # Stage 1: Training with collecting stats.
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, _, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size, rng=input_rng
    )

    trained_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), state.model
    )
    dtype = jnp.dtype
    expected_trained_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "WeightedStatsCalibration_0": {
                    "max_of_abs_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_l1_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_lp_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_ones": (dtype("float32"), (1, 1, 1)),
                    "sum_of_vals": (dtype("float32"), (1, 1, 1)),
                },
                "WeightedStatsCalibration_1": {
                    "max_of_abs_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_l1_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_lp_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_ones": (dtype("float32"), (2, 1, 10)),
                    "sum_of_vals": (dtype("float32"), (2, 1, 10)),
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "WeightedStatsCalibration_0": {
                    "max_of_abs_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_l1_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_lp_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_ones": (dtype("float32"), (1, 1, 1)),
                    "sum_of_vals": (dtype("float32"), (1, 1, 1)),
                },
                "WeightedStatsCalibration_1": {
                    "max_of_abs_vals": (dtype("float32"), (2, 1, 256)),
                    "sum_of_l1_vals": (dtype("float32"), (2, 1, 256)),
                    "sum_of_lp_vals": (dtype("float32"), (2, 1, 256)),
                    "sum_of_ones": (dtype("float32"), (2, 1, 256)),
                    "sum_of_vals": (dtype("float32"), (2, 1, 256)),
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "WeightedStatsCalibration_0": {
                    "max_of_abs_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_l1_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_lp_vals": (dtype("float32"), (1, 1, 1)),
                    "sum_of_ones": (dtype("float32"), (1, 1, 1)),
                    "sum_of_vals": (dtype("float32"), (1, 1, 1)),
                },
                "WeightedStatsCalibration_1": {
                    "max_of_abs_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_l1_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_lp_vals": (dtype("float32"), (2, 1, 10)),
                    "sum_of_ones": (dtype("float32"), (2, 1, 10)),
                    "sum_of_vals": (dtype("float32"), (2, 1, 10)),
                },
            }
        },
    }

    utils.test_pprint_eq(expected_trained_pytree, trained_pytree["qc"])

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    logits_before_conversion, params = forward(
        state.model, state.cnn_eval.apply
    )
    state = state.replace(model=params)

    # Stage 2. Convert the checkpoint, and serve.
    serve_fn, model_serving = flax_e2e_model.serving_conversion(
        state, weight_only=False
    )
    dtype = jnp.dtype
    expected_dtype = dtype("int4") if bits == 4 else dtype("int8")
    expected_aqt_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 5, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 1568, 256)),
                        scale=[(dtype("float32"), (2, 1, 256))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 128, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    utils.test_pprint_eq(expected_aqt_pytree, serving_pytree["aqt"])

    # Compare logits of models before conversion and after conversion.
    logits_after_conversion, _ = forward(model_serving, serve_fn)
    assert (logits_before_conversion == logits_after_conversion).all()

  @parameterized.parameters(
      (["e4m3"] * 2 + ["e5m2"] * 4,),  # Higher precision fwd, larger range bwd
      ([8] * 6,),
  )
  def test_mnist_delayed_scaling_calibration(self, bits):
    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng, label_rng = jax.random.split(rng, 3)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 64
    batch_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)
    num_train_steps = ds_size // batch_size

    # AQT Config
    aqt_cfg_dg = config.config_v4()
    aqt_cfg_dg = config.set_bits(aqt_cfg_dg, *bits)

    # Update cfg with DelayedScalingCalibration.
    # Setting history length to be greater than the number of training steps
    # so we can verify history is being updated correctly on each step below.
    amax_history_length = num_train_steps * 2
    calibration_cls = functools.partial(
        delayed_scaling_calibration.DelayedScalingCalibration,
        amax_history_length=amax_history_length,
    )
    aqt_cfg_dg.fwd.dg_quantizer.lhs.calibration = calibration_cls
    aqt_cfg_dg.fwd.dg_quantizer.rhs.calibration = calibration_cls
    aqt_cfg_dg.dlhs.dg_quantizer.lhs.calibration = calibration_cls
    aqt_cfg_dg.dlhs.dg_quantizer.rhs.calibration = calibration_cls
    aqt_cfg_dg.drhs.dg_quantizer.lhs.calibration = calibration_cls
    aqt_cfg_dg.drhs.dg_quantizer.rhs.calibration = calibration_cls

    # Stage 1: Training with Delayed Scaling Calibration
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, _, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size, rng=input_rng
    )

    trained_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape),
        state.model[delayed_scaling_calibration.CALIBRATION_STATS],
    )
    dtype = jnp.dtype
    expected_trained_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "DelayedScalingCalibration_0": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_1": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_2": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_3": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_4": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_5": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "DelayedScalingCalibration_0": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_1": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_2": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_3": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_4": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_5": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "DelayedScalingCalibration_0": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_1": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_2": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_3": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_4": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
                "DelayedScalingCalibration_5": {
                    "amax_history": (dtype("float32"), (amax_history_length,)),
                    "bound": (dtype("float32"), (1,)),
                },
            }
        },
    }

    utils.test_pprint_eq(expected_trained_pytree, trained_pytree)

    # Verify we get the number of updates we expect to amax_history. This is a
    # bit difficult to verify since sometimes the amax_update is 0, but a few of
    # of the modules have more predictable behavior

    # We should have num_train_steps + 1 (for init) updates for the forward pass
    self.assertEqual(
        jnp.sum(
            state.model[delayed_scaling_calibration.CALIBRATION_STATS][
                "Dense_0"
            ]["AqtDotGeneral_0"]["DelayedScalingCalibration_0"]["amax_history"][
                :
            ]
            != 0
        ).item(),
        num_train_steps + 1,
    )
    self.assertNotEqual(
        state.model[delayed_scaling_calibration.CALIBRATION_STATS]["Dense_0"][
            "AqtDotGeneral_0"
        ]["DelayedScalingCalibration_0"]["bound"][:].item(),
        1.0,
    )

    # We should have num_train_steps updates for the backward pass
    self.assertEqual(
        jnp.sum(
            state.model[delayed_scaling_calibration.CALIBRATION_STATS][
                "Dense_0"
            ]["AqtDotGeneral_0"]["DelayedScalingCalibration_3"]["amax_history"][
                :
            ]
            != 0
        ).item(),
        num_train_steps,
    )
    self.assertNotEqual(
        state.model[delayed_scaling_calibration.CALIBRATION_STATS]["Dense_0"][
            "AqtDotGeneral_0"
        ]["DelayedScalingCalibration_3"]["bound"][:].item(),
        1.0,
    )

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    # We need to set the mode to SERVE so the MutableArrays don't get changed,
    # which would mean the logits would not match before and after conversion.
    state.cnn_eval.weights_quant_mode = utils.QuantMode.SERVE
    state.cnn_eval.activations_quant_mode = utils.QuantMode.SERVE
    logits_before_conversion, params = forward(
        state.model, state.cnn_eval.apply
    )
    state = state.replace(model=params)

    # Stage 2. Convert the checkpoint, and serve.
    serve_fn, model_serving = flax_e2e_model.serving_conversion(
        state, weight_only=False
    )
    dtype = jnp.dtype
    expected_dtype = dtype("int8") if bits[0] == 8 else dtype("float8_e4m3fn")
    expected_aqt_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 5, 10)),
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 1568, 256)),
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=None,
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 128, 10)),
                        scale=[(dtype("float32"), (1, 1, 1))],
                        scale_t=None,
                        dequant_dtype=dtype("float32"),
                        bias=[],
                    )
                },
            }
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    utils.test_pprint_eq(expected_aqt_pytree, serving_pytree["aqt"])

    # Compare logits of models before conversion and after conversion.
    logits_after_conversion, _ = forward(model_serving, serve_fn)
    assert (logits_before_conversion == logits_after_conversion).all()

  def test_mnist_training_backward_compatibility(self):
    aqt_cfg_dg = config.config_v4()

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)

    # Stage 1: regular training
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, _, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size=ds_size // 2, rng=input_rng
    )

    # Run forward once more in the same mode to get logits for testing below.
    logits_s1, _ = forward(state.model, state.cnn_eval.apply)

    # Stage 2: Model conversion (quantized weights freezing)
    # Freeze with legacy freezer; serve with new freezer.
    apply_serving, model_serving = flax_e2e_model.serving_conversion(
        state, legacy_for_freeze=True, legacy_for_serve=False
    )

    dtype = jnp.dtype
    expected_aqt_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "scale": (dtype("float32"), (2, 1, 10)),
                    "value": (dtype("int8"), (2, 5, 10)),
                }
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "qrhs": {
                    "scale": (dtype("float32"), (2, 1, 256)),
                    "value": (dtype("int8"), (2, 1568, 256)),
                }
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "qrhs": {
                    "scale": (dtype("float32"), (2, 1, 10)),
                    "value": (dtype("int8"), (2, 128, 10)),
                }
            }
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    utils.test_pprint_eq(expected_aqt_pytree, serving_pytree["aqt"])

    # Stage 3: inference mode.
    logits_s3, _ = forward(model_serving, apply_serving)
    assert (logits_s3 == logits_s1).all()


if __name__ == "__main__":
  absltest.main()
